#IFNDEF __FB_UNIX__
  #DEFINE dSINGLE
  #LIBPATH "C:\opt\Gtk\lib"
#ELSE
  #DEFINE dDOUBLE
#ENDIF

#INCLUDE ONCE "ode/ode.bi"
#INCLUDE ONCE "../OpenGl_Player.bi"

#DEFINE MASSE 1
' define your variables here
STATIC SHARED AS dWorldID WORLD
STATIC SHARED AS dSpaceID SPACE_
STATIC SHARED AS dJointGroupID CONTACTGROUP
STATIC SHARED AS dBodyID BODY(2)
STATIC SHARED AS dGeomID GEOM(2)
STATIC SHARED AS dReal SIZE(2) => {0.3 ,0.2, 0.5}
STATIC SHARED AS dReal PAUSE_

#MACRO CALC_MATRIX(Geo)
' get position and rotation
  DIM AS CONST dReal PTR Po = dGeomGetPosition(GEOM(Geo))
  DIM AS CONST dReal PTR Ro = dGeomGetRotation(GEOM(Geo))
' coordinate transformation ODE->GL
  DIM AS GLfloat matrix(15)
  matrix(0) = Ro[0]
  matrix(1) = Ro[8]
  matrix(2) = Ro[4]
  matrix(3) = 0
  matrix(4) = Ro[2]
  matrix(5) = Ro[10]
  matrix(6) = Ro[6]
  matrix(7) = 0
  matrix(8) = Ro[1]
  matrix(9) = Ro[9]
  matrix(10) = Ro[5]
  matrix(11) = 0
  matrix(12) = Po[0]
  matrix(13) = Po[2]
  matrix(14) = Po[1]
  matrix(15) = 1
  glMultMatrixf(@matrix(0))
#ENDMACRO

SUB drawCube(BYVAL Geo AS INTEGER)
  STATIC AS dVector3 Sides
  glPushMatrix()
  CALC_MATRIX(Geo)
  dGeomBoxGetLengths(GEOM(Geo), @Sides)
  glScaled(Sides.x(0), Sides.x(2), Sides.x(1))
  gdk_gl_draw_cube(TRUE, 1.0)
  glPopMatrix()
END SUB

SUB drawSphere(BYVAL Geo AS INTEGER)
  STATIC AS dReal radius
  glPushMatrix()
  CALC_MATRIX(Geo)
  radius = dGeomSphereGetRadius(GEOM(Geo))
  glScaled(radius, radius, radius)
  gdk_gl_draw_sphere(TRUE, 1.0, 21, 12)
  glPopMatrix()
END SUB

' stuff to start the simulation
SUB new_sim CDECL(BYREF Ti AS guint32) STATIC
  PAUSE_ = 0.001 * Ti

  dInitODE()

' create .WORLD
  WORLD = dWorldCreate()
  SPACE_ = dHashSpaceCreate(0)
  dWorldSetGravity(WORLD, 0, 0, -0.981)
  dWorldSetCFM(WORLD, 1e-5)
  dCreatePlane(SPACE_, 0, 0, 1, 0)
  CONTACTGROUP = dJointGroupCreate(0)

  STATIC AS dMass mass
' create object (upper ball)
  BODY(0) = dBodyCreate(WORLD)
  GEOM(0) = dCreateSphere(SPACE_, SIZE(0))
  dMassSetSphere(@mass, MASSE, SIZE(0))
  dBodySetMass(BODY(0), @mass)
  dGeomSetBody(GEOM(0), BODY(0))
  dBodySetPosition(BODY(0), 0.0, 0.0, 3.5)

' create object (lower ball)
  BODY(1) = dBodyCreate(WORLD)
  GEOM(1) = dCreateSphere(SPACE_, SIZE(1))
  dMassSetSphere(@mass, 0.5 * MASSE, SIZE(1))
  dBodySetMass(BODY(1), @mass)
  dGeomSetBody(GEOM(1), BODY(1))
  dBodySetPosition(BODY(1), 0.05, 0.0, 1.9)

' create object (box)
  BODY(2) = dBodyCreate(WORLD)
  GEOM(2) = dCreateBox(SPACE_, SIZE(2), SIZE(2), SIZE(2))
  dMassSetBox(@mass, 2 * MASSE, SIZE(2), SIZE(2), SIZE(2))
  dBodySetMass(BODY(2), @mass)
  dGeomSetBody(GEOM(2), BODY(2))
  dBodySetPosition(BODY(2), 0.25 * SIZE(1), 0.0, 0.5 * SIZE(2))

' init OpenGL
  glPushAttrib(GL_LIGHTING_BIT)
  glLoadIdentity()

  STATIC AS GLfloat ambient(...) = { 0.223529, 0.329412, 0.027451, 1.0 }
  STATIC AS GLfloat diffuse(...) = { 0.568627, 0.780392, 0.113725, 1.0 }
  STATIC AS GLfloat position(...) = { -2.0, 5.0, 11.0, 0.0 }
  glLightfv(GL_LIGHT0, GL_AMBIENT, @ambient(0))
  glLightfv(GL_LIGHT0, GL_DIFFUSE, @diffuse(0))
  glLightfv(GL_LIGHT0, GL_POSITION, @position(0))

  STATIC AS GLfloat lmodel_ambient(...) = { 0.2, 0.2, 0.2, 1.0 }
  STATIC AS GLfloat local_view = 0.0
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @lmodel_ambient(0))
  glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, @local_view)

  STATIC AS GLfloat specular(...) = { 0.1992157, 0.1941176, 0.1807843, 1.0 }
  STATIC AS GLfloat shininess = 0.21794872 * 128.0
  'glMaterialfv(GL_FRONT, GL_SPECULAR, @specular(0))
  glMaterialf(GL_FRONT, GL_SHININESS, shininess)

  glEnable(GL_LIGHTING)
  glEnable(GL_LIGHT0)
  glEnable(GL_DEPTH_TEST)

  glClearColor(1.0, 1.0, 1.0, 1.0)
  glClearDepth(1.0)
END SUB

' stuff to end the simulation
SUB del_sim CDECL(BYREF Ti AS guint32) STATIC
  glPopAttrib()
  dJointGroupDestroy(CONTACTGROUP)
  dWorldDestroy(WORLD)
  dCloseODE()
END SUB

' stuff GtkWidget::expose event for the OpenGL widget (drawing)
SUB expose CDECL() STATIC
  glClear(GL_COLOR_BUFFER_BIT OR GL_DEPTH_BUFFER_BIT)
  glLoadIdentity()

  glTranslatef(0.0, -1.50, -11.0) '                      camera position
  glRotatef(-9, 0, 1, 0) '                                 camera ankle

  drawSphere(0)
  drawSphere(1)
  drawCube(2)
END SUB

'this is called by dSpaceCollide when two objects in space are
'potentially colliding.
SUB nearCallback CDECL(BYVAL Poi AS ANY PTR, BYVAL G1 AS dGeomID, BYVAL G2 AS dGeomID) STATIC
  STATIC AS dBodyID b1, b2
  STATIC AS dJointID joint
  STATIC AS dContact contact

  b1 = dGeomGetBody(G1)
  b2 = dGeomGetBody(G2)
  IF dCollide(G1, G2, 1, @contact.GEOM, SIZEOF(dContact)) THEN
    contact.surface.mode = dContactBounce OR dContactSoftCFM
    contact.surface.mu = dInfinity '                    friction parameter
    contact.surface.bounce = 0.75 '   bounce is the amount of "bouncyness"
    contact.surface.bounce_vel = 0.1 '     min. velocity to cause a bounce
    contact.surface.soft_cfm = 0.001 '   constraint force mixing parameter
    joint = dJointCreateContact(WORLD, CONTACTGROUP, @contact)
    dJointAttach(joint, b1, b2)
  END IF
END SUB

' stuff for simulation loop
SUB time_out CDECL() STATIC
  ' find collisions and add contact joints
  dSpaceCollide(SPACE_, NULL, @nearCallback())
  ' step the simulation
  dWorldQuickStep(WORLD, PAUSE_)
  ' remove all contact joints
  dJointGroupEmpty(CONTACTGROUP)
END SUB

'' stuff to react on keypress
'FUNCTION keyboard CDECL(BYVAL Ev AS GdkEventKey PTR) AS gboolean STATIC
  '' handle Ev here, return FALSE for default Ev handling, TRUE if you handled Ev
  'RETURN FALSE
'END FUNCTION

'' stuff for setting parameters in a GTK window
'SUB parameters CDECL() STATIC
'END SUB

' stuff to report for Sim_Player
SUB sim_init CDECL ALIAS "sim_init"(BYVAL I AS _Interface PTR) STATIC EXPORT
  I->new_sim = @new_sim()
  I->del_sim = @del_sim() ' or NULL
  I->expos = @expose() ' or NULL
  I->keybo = NULL '@keyboard() ' or NULL
  I->param = NULL '@parameters() ' or NULL
  I->tiout = @time_out()
  I->descr = "Two balls bouncing on a box"
END SUB

