#IFNDEF __FB_UNIX__
  #DEFINE dSINGLE
  #LIBPATH "C:\opt\Gtk\lib"
#ELSE
  #DEFINE dDOUBLE
#ENDIF

#INCLUDE ONCE "ode/ode.bi"
#INCLUDE ONCE "../OpenGl_Player.bi"

#MACRO CALC_MATRIX(_I_)
' get position and rotation
  STATIC AS CONST dReal PTR Po, Ro
  Po = dGeomGetPosition(GEOM(_I_))
  Ro = dGeomGetRotation(GEOM(_I_))
' coordinate transformation ODE->GL
  STATIC AS GLfloat matrix(15)
  matrix(0) = Ro[0]
  matrix(1) = Ro[8]
  matrix(2) = Ro[4]
  matrix(3) = 0
  matrix(4) = Ro[2]
  matrix(5) = Ro[10]
  matrix(6) = Ro[6]
  matrix(7) = 0
  matrix(8) = Ro[1]
  matrix(9) = Ro[9]
  matrix(10) = Ro[5]
  matrix(11) = 0
  matrix(12) = Po[0]
  matrix(13) = Po[2]
  matrix(14) = Po[1]
  matrix(15) = 1
  glMultMatrixf(@matrix(0))
#ENDMACRO


STATIC SHARED AS INTEGER ROWS = 6, COLS = 6, AZ
STATIC SHARED AS dReal DIST = 21.0
STATIC SHARED AS dWorldID WORLD
STATIC SHARED AS dSpaceID SPACE_
STATIC SHARED AS dReal PAUSE_
STATIC SHARED AS dReal SIZE = 0.5
STATIC SHARED AS dReal MAS = 4
STATIC SHARED AS dJointGroupID CONTACTGROUP
STATIC SHARED AS dBodyID BODY()
STATIC SHARED AS dGeomID GEOM()

SUB drawCube CDECL(BYVAL Geo AS INTEGER) STATIC
  STATIC AS dVector3 Sides
  glPushMatrix()
  CALC_MATRIX(Geo)
  dGeomBoxGetLengths(GEOM(Geo), @Sides)
  glScaled(Sides.x(0), Sides.x(2), Sides.x(1))
  gdk_gl_draw_cube(TRUE, 1.0)
  glPopMatrix()
END SUB

' stuff to start the simulation
SUB new_sim CDECL(BYREF Ti AS guint32) STATIC
  PAUSE_ = 0.001 * Ti

  dInitODE()

' create .WORLD
  WORLD = dWorldCreate()
  SPACE_ = dHashSpaceCreate(0)
  dWorldSetGravity(WORLD, 0, 0, -0.981)
  dWorldSetCFM(WORLD, 1e-5)
  dCreatePlane(SPACE_, 0, 0, 1, 0)
  CONTACTGROUP = dJointGroupCreate(0)

  STATIC AS dMass mass
  STATIC AS dReal yo = -2.0, zo, x = 0, y, z, n
  AZ = ROWS * COLS - 1 : zo = 1.5 * SIZE : n = 0
  REDIM BODY(AZ)
  REDIM GEOM(AZ)
  FOR r AS INTEGER = 1 TO COLS
    x = -4.0 + SIZE * 1.3 * r
    z = zo
    FOR c AS INTEGER = 1 TO ROWS
      BODY(n) = dBodyCreate(WORLD)
      GEOM(n) = dCreateBox(SPACE_, SIZE, SIZE, SIZE)
      dMassSetBox(@mass, MAS, SIZE, SIZE, SIZE)
      dBodySetMass(BODY(n), @mass)
      dGeomSetBody(GEOM(n), BODY(n))
      dBodySetPosition(BODY(n), x, y, z)
      x += .65 * SIZE
      z += SIZE
      n += 1
    NEXT
  NEXT

' set OpenGl parameters
  glPushAttrib(GL_LIGHTING_BIT)
  glLoadIdentity()
  STATIC AS GLfloat ambient(...) = { 0.329412, 0.223529, 0.027451, 1.0 }
  STATIC AS GLfloat diffuse(...) = { 0.780392, 0.568627, 0.113725, 1.0 }
  STATIC AS GLfloat position(0 TO 3) = { 1.0, 2.0, 11.0, 0.0 }
  glLightfv(GL_LIGHT0, GL_AMBIENT, @ambient(0))
  glLightfv(GL_LIGHT0, GL_DIFFUSE, @diffuse(0))
  glLightfv(GL_LIGHT0, GL_POSITION, @position(0))

  STATIC AS GLfloat lmodel_ambient(0 TO 3) = { 0.2, 0.2, 0.2, 1.0 }
  STATIC AS GLfloat local_view(0) = { 0.0 }
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @lmodel_ambient(0))
  glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, @local_view(0))
  glEnable(GL_LIGHTING)
  glEnable(GL_LIGHT0)
  glEnable(GL_DEPTH_TEST)

  glClearColor(1.0, 1.0, 1.0, 1.0)
  glClearDepth(1.0)
END SUB

' stuff to end the simulation
SUB del_sim CDECL(BYREF Ti AS guint32) STATIC
  glPopAttrib()
  dJointGroupDestroy(CONTACTGROUP)
  dWorldDestroy(WORLD)
  dCloseODE()
END SUB

' stuff GtkWidget::expose event for the OpenGL widget
SUB expose CDECL() STATIC
  glClear(GL_COLOR_BUFFER_BIT OR GL_DEPTH_BUFFER_BIT)
  glLoadIdentity()
  glTranslatef( -1.0, -1.50, -DIST) '                    camera position
  glRotatef(-17, 0., 1., 0.) '                           camera rotation
  glRotatef( 15, 1., 0., 0.) '                           camera rotation

  FOR x AS INTEGER = 0 TO AZ
    drawCube(x)
  NEXT
END SUB

SUB nearCallback CDECL(BYVAL Poi AS ANY PTR, BYVAL G1 AS dGeomID, BYVAL G2 AS dGeomID) STATIC
  STATIC AS dBodyID b1, b2
  STATIC AS dJointID c
  STATIC AS dContact contact

  b1 = dGeomGetBody(G1)
  b2 = dGeomGetBody(G2)
  contact.surface.mode = dContactBounce OR dContactSoftCFM
  contact.surface.mu = dInfinity '                    friction parameter
  contact.surface.bounce = 0.75 '   bounce is the amount of "bouncyness"
  contact.surface.bounce_vel = 0.1 '     min. velocity to cause a bounce
  contact.surface.soft_cfm = 0.001 '   constraint force mixing parameter
  IF dCollide(G1, G2, 1, @contact.GEOM, SIZEOF(dContact)) THEN
    c = dJointCreateContact(WORLD, CONTACTGROUP, @contact)
    dJointAttach(c, b1, b2)
  END IF
END SUB

' stuff for simulation loop
SUB time_out CDECL() STATIC
  ' find collisions and add contact joints
  dSpaceCollide(SPACE_, NULL, @nearCallback())
  ' step the simulation
  dWorldQuickStep(WORLD, PAUSE_)
  ' remove all contact joints
  dJointGroupEmpty(CONTACTGROUP)
END SUB

' stuff to react on keypress
FUNCTION keyboard CDECL(BYVAL Ev AS GdkEventKey PTR) AS gboolean STATIC
  ' handle Ev here, return TRUE to stop further handling (or FALSE)

  RETURN FALSE
END FUNCTION

#INCLUDE ONCE "ODE_Boxes_tobac.bas"
' stuff for setting parameters in a GTK window
SUB parameters CDECL(BYVAL Wid AS GtkWindow PTR) STATIC

  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_Rows), ROWS)
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_Cols), COLS)
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_Dist), DIST)
  IF gtk_dialog_run(GTK_DIALOG(GUI.Dial_Para)) THEN
    DIM AS guint32 Ti
    del_sim(Ti)
    ROWS = CUINT(gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_Rows)))
    COLS = CUINT(gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_Cols)))
    DIST = gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_Dist))
    new_sim(PAUSE_ * 1000)
  END IF
  gtk_widget_hide(GTK_WIDGET(GUI.Dial_Para))


END SUB

' stuff to report for Sim_Player
SUB sim_init CDECL ALIAS "sim_init"(BYVAL I AS _Interface PTR) STATIC EXPORT
  I->new_sim = @new_sim()
  I->del_sim = @del_sim() ' or NULL
  I->expos = @expose()
  I->keybo = @keyboard() ' or NULL
  I->param = @parameters() ' or NULL
  I->tiout = @time_out()
  I->descr = "A stack of boxes falling"
END SUB

