#IFNDEF __FB_UNIX__
  #DEFINE dSINGLE
  #LIBPATH "C:\opt\Gtk\lib"
#ELSE
  #DEFINE dDOUBLE
#ENDIF

#INCLUDE ONCE "ode/ode.bi"
#INCLUDE ONCE "../OpenGl_Player.bi"

#MACRO CALC_MATRIX(_I_)
' get position and rotation
  STATIC AS CONST dReal PTR Po, Ro
  Po = dGeomGetPosition(Geom(_I_))
  Ro = dGeomGetRotation(Geom(_I_))
' coordinate transformation ODE->GL
  STATIC AS GLfloat matrix(15)
  matrix(0) = Ro[0]
  matrix(1) = Ro[8]
  matrix(2) = Ro[4]
  matrix(3) = 0
  matrix(4) = Ro[2]
  matrix(5) = Ro[10]
  matrix(6) = Ro[6]
  matrix(7) = 0
  matrix(8) = Ro[1]
  matrix(9) = Ro[9]
  matrix(10) = Ro[5]
  matrix(11) = 0
  matrix(12) = Po[0]
  matrix(13) = Po[2]
  matrix(14) = Po[1]
  matrix(15) = 1
  glMultMatrixf(@matrix(0))
#ENDMACRO

STATIC SHARED AS INTEGER AZ = 99
STATIC SHARED AS dReal DOM_DIST = 0.63
STATIC SHARED AS dReal DOM_RAD = 1.7
STATIC SHARED AS dReal CAM_DIST = 24

STATIC SHARED AS dWorldID World
STATIC SHARED AS dSpaceID Space_
STATIC SHARED AS dJointGroupID Contactgroup
STATIC SHARED AS dReal Pause_
STATIC SHARED AS dMass Mass
STATIC SHARED AS dReal mas = 4
STATIC SHARED AS dBodyID Ground
STATIC SHARED AS dBodyID Body()
STATIC SHARED AS dGeomID Geom()
STATIC SHARED AS dReal Size(2) = {0.1, 0.3, 1.0}

SUB drawCube CDECL(BYVAL Geo AS INTEGER) STATIC
  STATIC AS dVector3 Sides
  glPushMatrix()
  CALC_MATRIX(Geo)
  dGeomBoxGetLengths(Geom(Geo), @Sides)
  glScaled(Sides.x(0), Sides.x(2), Sides.x(1))
  gdk_gl_draw_cube(TRUE, 1.0)
  glPopMatrix()
END SUB

' stuff GtkWidget::expose event for the OpenGL widget
SUB expos CDECL() STATIC
  glClear(GL_COLOR_BUFFER_BIT OR GL_DEPTH_BUFFER_BIT)
  glLoadIdentity()
  glTranslatef( 0.0, 0.00, -CAM_DIST) '                    camera position
  glRotatef(50, 1, 0, 0) '                                  camera ankle

  FOR x AS INTEGER = 0 TO AZ
    drawCube(x)
  NEXT
END SUB

'this is called by dSpaceCollide when two objects in Space_ are
'potentially colliding.
SUB nearCallback CDECL(BYVAL Poi AS ANY PTR, _
                       BYVAL G1 AS dGeomID, BYVAL G2 AS dGeomID) STATIC
  STATIC AS dBodyID b1, b2
  STATIC AS dJointID c
  STATIC AS dContact contact

  b1 = dGeomGetBody(G1)
  b2 = dGeomGetBody(G2)
  IF dCollide(G1, G2, 1, @contact.geom, SIZEOF(dContact)) THEN
    IF G1 <> Ground ANDALSO G2 <> Ground THEN
      contact.surface.mu = 0.2 '                      friction parameter
      contact.surface.mode = dContactBounce
      contact.surface.bounce = 0.95 '    bounce (amount of "bouncyness")
      contact.surface.bounce_vel = 0.01 ' min velocity to cause a bounce
    ELSE
      contact.surface.mode = dContactBounce OR dContactSoftCFM
      contact.surface.mu = dInfinity '                friction parameter
      contact.surface.bounce = 0.25 '    bounce (amount of "bouncyness")
      contact.surface.bounce_vel = 0.95 ' min velocity to cause a bounce
    END IF
    c = dJointCreateContact(World, Contactgroup, @contact)
    dJointAttach(c, b1, b2)
  END IF
END SUB

' stuff for simulation loop
SUB tiout CDECL() STATIC
  dSpaceCollide(Space_, NULL, @nearCallback()) ' find collisions and add contact joints
  dWorldQuickStep(World, Pause_) '                   step the simulation
  dJointGroupEmpty(Contactgroup) '             remove all contact joints
END SUB

SUB new_sim CDECL(BYREF Ti AS guint32) STATIC
  Pause_ = 0.001 * Ti

  dInitODE()

' create .World
  World = dWorldCreate()
  Space_ = dHashSpaceCreate(0)
  dWorldSetGravity(World, 0, 0, -0.981)
  dWorldSetCFM(World, 1e-10)
  Ground = dCreatePlane(Space_, 0, 0, 1, 0)
  Contactgroup = dJointGroupCreate(0)

  STATIC AS dReal z, ds, wi, r, x, y
  z = .5 * Size(2) : ds = DOM_DIST * Size(2) : wi = 0.0
  STATIC AS dQuaternion rot
  REDIM Body(AZ)
  REDIM Geom(AZ)
  FOR n AS INTEGER = 0 TO AZ
    Body(n) = dBodyCreate(World)
    Geom(n) = dCreateBox(Space_, Size(0), Size(1), Size(2))
    dMassSetBox(@Mass, mas, Size(0), Size(1), Size(2))
    dBodySetMass(Body(n), @Mass)
    dGeomSetBody(Geom(n), Body(n))

    r = DOM_RAD + wi / M_Pi / 2
    x = r * SIN(wi)
    y = r * COS(wi)
    dBodySetPosition(Body(n), x, y, z)
    dQFromAxisAndAngle(@rot, 0.0, 0.0, 1.0, -wi)
    dBodySetQuaternion(Body(n), @rot)
    wi += ATN(ds / r)
  NEXT
  dBodyAddRelForceAtRelPos(Body(0), 3., 0., 0., 0., 0., z)

' set OpenGL parameters
  glPushAttrib(GL_LIGHTING_BIT)
  glLoadIdentity()
  STATIC AS GLfloat ambient(...) = { 0.7, 0.7, 0.7, 1.0 }
  STATIC AS GLfloat diffuse(...) = { 1.0, 1.0, 1.0, 1.0 }
  STATIC AS GLfloat position(...) = { -0.0, 1.0, 0.0, 1.0 }
  glLightfv(GL_LIGHT1, GL_AMBIENT, @ambient(0))
  glLightfv(GL_LIGHT1, GL_DIFFUSE, @diffuse(0))
  glLightfv(GL_LIGHT1, GL_POSITION, @position(0))

  STATIC AS GLfloat lmodel_ambient(...) = { 0.5, 0.5, 0.5, 1.0 }
  STATIC AS GLfloat local_view = 1.0
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @lmodel_ambient(0))
  glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, @local_view)

  STATIC AS Glfloat mat_AMBIENT(...) = {0.0, 0.0, 0.0, 1.0}
  STATIC AS Glfloat mat_DIFFUS(...) = { 0.01, 0.01, 0.01, 1.0}
  STATIC AS Glfloat mat_SPECULAR(...) = { 0.50, 0.50, 0.50, 1.0}
  glMaterialfv(GL_FRONT, GL_AMBIENT, @mat_AMBIENT(0))
  glMaterialfv(GL_FRONT, GL_DIFFUSE, @mat_DIFFUS(0))
  glMaterialfv(GL_FRONT, GL_SPECULAR, @mat_SPECULAR(0))
  glMaterialf(GL_FRONT, GL_SHININESS, 32.)
  glEnable(GL_LIGHTING)
  glEnable(GL_LIGHT1)
  glEnable(GL_DEPTH_TEST)

  glClearColor(0.5, 0.79, 0.95, 1.0)
  glClearDepth(1.0)
END SUB

SUB del_sim CDECL(BYREF Ti AS guint32) STATIC
  glPopAttrib()
  dJointGroupDestroy(Contactgroup)
  dWorldDestroy(World)
  dCloseODE()
END SUB

' stuff to set the variable parameters in a GTK window
#INCLUDE ONCE "ODE_Dominoes_tobac.bas"
SUB parameters CDECL(BYVAL Wid AS GtkWindow PTR) STATIC
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_Az), AZ + 1)
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_DomDist), DOM_DIST * 100)
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_DomRad), DOM_Rad * 100)
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GUI.Adj_ViewDist), CAM_DIST)
  IF gtk_dialog_run(GTK_DIALOG(GUI.Dial_Para)) THEN
    DIM AS guint32 Ti
    del_sim(Ti)
    AZ = CUINT(gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_Az))) - 1
    DOM_DIST = gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_DomDist)) * .01
    DOM_RAD = gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_DomRad)) * .01
    CAM_DIST = gtk_adjustment_get_value(GTK_ADJUSTMENT(GUI.Adj_ViewDist))
    new_sim(PAUSE_ * 1000)
  END IF
  gtk_widget_hide(GTK_WIDGET(GUI.Dial_Para))
END SUB

SUB sim_init CDECL ALIAS "sim_init"(BYVAL I AS _Interface PTR) STATIC EXPORT
  I->new_sim = @new_sim()
  I->del_sim = @del_sim()
  I->expos = @expos()
  I->keybo = NULL
  I->param = @parameters()
  I->tiout = @tiout()
  i->descr = "Simulate dominoes with ODE"
END SUB


